==> mastodon_api-0.0.0.tar.gz: lib/src/core/service_helper.dart <==
// Copyright 2022 Kato Shinya. All rights reserved.
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided the conditions.

// Dart imports:
import 'dart:convert' as converter;

// Package imports:
import 'package:http/http.dart' as http;
import 'package:http/http.dart';

// Project imports:
import 'client/client_context.dart';
import 'client/stream_request.dart';
import 'client/stream_response.dart';
import 'serializable.dart';
import 'util/json_utils.dart';

abstract class Service {
  Future<http.Response> get(
    String unencodedPath, {
    Response Function(Response response)? validate,
  });

  Future<StreamResponse> getStream(
    final String unencodedPath, {
    Map<String, dynamic> queryParameters = const {},
    Map<String, dynamic> Function(
            StreamedResponse streamedResponse, String event)?
        validate,
  });

  Future<http.Response> post(
    String unencodedPath, {
    Map<String, dynamic> queryParameters = const {},
    Map<String, String> body = const {},
    Response Function(Response response)? validate,
  });

  Future<http.Response> postMultipart(
    String unencodedPath, {
    List<http.MultipartFile> files = const [],
    Map<String, dynamic> queryParameters = const {},
    Response Function(Response response)? validate,
  });

  Future<http.Response> delete(
    String unencodedPath, {
    Response Function(Response response)? validate,
  });

  Future<http.Response> put(
    String unencodedPath, {
    Map<String, String> body = const {},
    Response Function(Response response)? validate,
  });
}

class ServiceHelper implements Service {
  /// Returns the new instance of [ServiceHelper].
  const ServiceHelper({
    required String authority,
    required ClientContext context,
  })  : _authority = authority,
        _context = context;

  /// The base url
  final String _authority;

  /// The mastodon client
  final ClientContext _context;

  @override
  Future<http.Response> get(
    final String unencodedPath, {
    Map<String, dynamic> queryParameters = const {},
    Response Function(Response response)? validate,
  }) async {
    final response = await _context.get(
      Uri.https(
        _authority,
        unencodedPath,
        _convertQueryParameters(queryParameters),
      ),
    );

    return validate != null ? validate(response) : response;
  }

  @override
  Future<StreamResponse> getStream(
    final String unencodedPath, {
    Map<String, dynamic> queryParameters = const {},
    Map<String, dynamic> Function(
            StreamedResponse streamedResponse, String event)?
        validate,
  }) async {
    final streamedResponse = await _context.getStream(
      StreamRequest(
        Uri.https(
          _authority,
          unencodedPath,
          _convertQueryParameters(queryParameters),
        ),
      ),
    );

    return StreamResponse(
      headers: streamedResponse.headers,
      body: streamedResponse.stream.transform(converter.utf8.decoder).map(
            (event) => validate != null
                ? validate(streamedResponse, event)
                : tryJsonDecode(streamedResponse, event),
          ),
    );
  }

  @override
  Future<http.Response> post(
    final String unencodedPath, {
    Map<String, dynamic> queryParameters = const {},
    dynamic body = const {},
    Response Function(Response response)? validate,
  }) async {
    final response = await _context.post(
      Uri.https(
        _authority,
        unencodedPath,
        _convertQueryParameters(queryParameters),
      ),
      headers: {'Content-type': 'application/json'},
      body: converter.jsonEncode(_removeNullValues(body)),
    );

    return validate != null ? validate(response) : response;
  }

  @override
  Future<http.Response> postMultipart(
    final String unencodedPath, {
    List<http.MultipartFile> files = const [],
    Map<String, dynamic> queryParameters = const {},
    Response Function(Response response)? validate,
  }) async {
    final response = await _context.postMultipart(
      Uri.https(
        _authority,
        unencodedPath,
        _convertQueryParameters(queryParameters),
      ),
      files: files,
    );

    return validate != null ? validate(response) : response;
  }

  @override
  Future<http.Response> delete(
    final String unencodedPath, {
    Response Function(Response response)? validate,
  }) async {
    final response = await _context.delete(
      Uri.https(_authority, unencodedPath),
    );

    return validate != null ? validate(response) : response;
  }

  @override
  Future<http.Response> put(
    final String unencodedPath, {
    dynamic body = const {},
    Response Function(Response response)? validate,
  }) async {
    final response = await _context.put(
      Uri.https(_authority, unencodedPath),
      headers: {'Content-type': 'application/json'},
      body: converter.jsonEncode(_removeNullValues(body)),
    );

    return validate != null ? validate(response) : response;
  }

  dynamic _removeNullValues(final dynamic object) {
    if (object is Map) {
      final parameters = <String, dynamic>{};
      object.forEach((key, value) {
        final newObject = _removeNullValues(value);

        if (newObject != null) {
          parameters[key] = newObject;
        }
      });

      return parameters.isNotEmpty ? parameters : null;
    } else if (object is List) {
      final parameters = <dynamic>[];
      for (final value in object) {
        final newObject = _removeNullValues(value);

        if (newObject != null) {
          parameters.add(newObject);
        }
      }

      return parameters.isNotEmpty ? parameters : null;
    }

    //! Just return it as is if it's neither Map nor List.
    return object;
  }

  Map<String, String> _convertQueryParameters(
    final Map<String, dynamic> queryParameters,
  ) {
    final serializedParameters = queryParameters.map((key, value) {
      if (value is List<Serializable>?) {
        return MapEntry(
          key,
          value?.toSet().map((e) => e.value).toList().join(','),
        );
      } else if (value is List<Enum>?) {
        return MapEntry(
          key,
          value?.toSet().map((e) => e.name).join(','),
        );
      } else if (value is List?) {
        return MapEntry(
          key,
          value?.toSet().join(','),
        );
      } else if (value is Serializable) {
        return MapEntry(
          key,
          value.value,
        );
      }

      return MapEntry(key, value);
    });

    return Map.from(_removeNullValues(serializedParameters) ?? {}).map(
      //! Uri.https(...) needs iterable in the value for query params by
      //! which it means a String in the value of the Map too. So you need
      //! to convert it from Map<String, dynamic> to Map<String, String>
      (key, value) {
        if (value is DateTime) {
          return MapEntry(key, value.toUtc().toIso8601String());
        }

        return MapEntry(key, value.toString());
      },
    );
  }
}

==> mastodon_api-0.1.0.tar.gz: lib/src/core/service_helper.dart <==
// Copyright 2022 Kato Shinya. All rights reserved.
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided the conditions.

// Dart imports:
import 'dart:convert' as converter;

// Package imports:
import 'package:http/http.dart' as http;

// Project imports:
import 'client/client_context.dart';
import 'client/stream_request.dart';
import 'client/stream_response.dart';
import 'client/user_context.dart';
import 'serializable.dart';
import 'util/json_utils.dart';

abstract class Service {
  Future<http.Response> get(
    UserContext userContext,
    String unencodedPath, {
    http.Response Function(http.Response response)? validate,
  });

  Future<StreamResponse> getStream(
    UserContext userContext,
    final String unencodedPath, {
    Map<String, dynamic> queryParameters = const {},
    Map<String, dynamic> Function(
            http.StreamedResponse streamedResponse, String event)?
        validate,
  });

  Future<http.Response> post(
    UserContext userContext,
    String unencodedPath, {
    Map<String, dynamic> queryParameters = const {},
    Map<String, String> body = const {},
    http.Response Function(http.Response response)? validate,
  });

  Future<http.Response> postMultipart(
    UserContext userContext,
    String unencodedPath, {
    List<http.MultipartFile> files = const [],
    Map<String, dynamic> queryParameters = const {},
    http.Response Function(http.Response response)? validate,
  });

  Future<http.Response> delete(
    UserContext userContext,
    String unencodedPath, {
    http.Response Function(http.Response response)? validate,
  });

  Future<http.Response> put(
    UserContext userContext,
    String unencodedPath, {
    Map<String, String> body = const {},
    http.Response Function(http.Response response)? validate,
  });
}

class ServiceHelper implements Service {
  /// Returns the new instance of [ServiceHelper].
  const ServiceHelper({
    required String authority,
    required ClientContext context,
  })  : _authority = authority,
        _context = context;

  /// The base url
  final String _authority;

  /// The mastodon client
  final ClientContext _context;

  @override
  Future<http.Response> get(
    UserContext userContext,
    final String unencodedPath, {
    Map<String, String> headers = const {},
    Map<String, dynamic> queryParameters = const {},
    http.Response Function(http.Response response)? validate,
  }) async {
    final response = await _context.get(
      userContext,
      Uri.https(
        _authority,
        unencodedPath,
        _convertQueryParameters(queryParameters),
      ),
      headers: headers,
    );

    return validate != null ? validate(response) : response;
  }

  @override
  Future<StreamResponse> getStream(
    UserContext userContext,
    final String unencodedPath, {
    Map<String, dynamic> queryParameters = const {},
    Map<String, dynamic> Function(
            http.StreamedResponse streamedResponse, String event)?
        validate,
  }) async {
    final streamedResponse = await _context.getStream(
      userContext,
      StreamRequest(
        Uri.https(
          _authority,
          unencodedPath,
          _convertQueryParameters(queryParameters),
        ),
      ),
    );

    return StreamResponse(
      headers: streamedResponse.headers,
      body: streamedResponse.stream.transform(converter.utf8.decoder).map(
            (event) => validate != null
                ? validate(streamedResponse, event)
                : tryJsonDecode(streamedResponse, event),
          ),
    );
  }

  @override
  Future<http.Response> post(
    UserContext userContext,
    final String unencodedPath, {
    Map<String, dynamic> queryParameters = const {},
    dynamic body = const {},
    http.Response Function(http.Response response)? validate,
  }) async {
    final response = await _context.post(
      userContext,
      Uri.https(
        _authority,
        unencodedPath,
        _convertQueryParameters(queryParameters),
      ),
      headers: {'Content-type': 'application/json'},
      body: converter.jsonEncode(_removeNullValues(body)),
    );

    return validate != null ? validate(response) : response;
  }

  @override
  Future<http.Response> postMultipart(
    UserContext userContext,
    final String unencodedPath, {
    List<http.MultipartFile> files = const [],
    Map<String, dynamic> queryParameters = const {},
    http.Response Function(http.Response response)? validate,
  }) async {
    final response = await _context.postMultipart(
      userContext,
      Uri.https(
        _authority,
        unencodedPath,
        _convertQueryParameters(queryParameters),
      ),
      files: files,
    );

    return validate != null ? validate(response) : response;
  }

  @override
  Future<http.Response> delete(
    UserContext userContext,
    final String unencodedPath, {
    http.Response Function(http.Response response)? validate,
  }) async {
    final response = await _context.delete(
      userContext,
      Uri.https(_authority, unencodedPath),
    );

    return validate != null ? validate(response) : response;
  }

  @override
  Future<http.Response> put(
    UserContext userContext,
    final String unencodedPath, {
    dynamic body = const {},
    http.Response Function(http.Response response)? validate,
  }) async {
    final response = await _context.put(
      userContext,
      Uri.https(_authority, unencodedPath),
      headers: {'Content-type': 'application/json'},
      body: converter.jsonEncode(_removeNullValues(body)),
    );

    return validate != null ? validate(response) : response;
  }

  dynamic _removeNullValues(final dynamic object) {
    if (object is Map) {
      final parameters = <String, dynamic>{};
      object.forEach((key, value) {
        final newObject = _removeNullValues(value);

        if (newObject != null) {
          parameters[key] = newObject;
        }
      });

      return parameters.isNotEmpty ? parameters : null;
    } else if (object is List) {
      final parameters = <dynamic>[];
      for (final value in object) {
        final newObject = _removeNullValues(value);

        if (newObject != null) {
          parameters.add(newObject);
        }
      }

      return parameters.isNotEmpty ? parameters : null;
    }

    //! Just return it as is if it's neither Map nor List.
    return object;
  }

  Map<String, String> _convertQueryParameters(
    final Map<String, dynamic> queryParameters,
  ) {
    final serializedParameters = queryParameters.map((key, value) {
      if (value is List<Serializable>?) {
        return MapEntry(
          key,
          value?.toSet().map((e) => e.value).toList().join(','),
        );
      } else if (value is List<Enum>?) {
        return MapEntry(
          key,
          value?.toSet().map((e) => e.name).join(','),
        );
      } else if (value is List?) {
        return MapEntry(
          key,
          value?.toSet().join(','),
        );
      } else if (value is Serializable) {
        return MapEntry(
          key,
          value.value,
        );
      }

      return MapEntry(key, value);
    });

    return Map.from(_removeNullValues(serializedParameters) ?? {}).map(
      //! Uri.https(...) needs iterable in the value for query params by
      //! which it means a String in the value of the Map too. So you need
      //! to convert it from Map<String, dynamic> to Map<String, String>
      (key, value) {
        if (value is DateTime) {
          return MapEntry(key, value.toUtc().toIso8601String());
        }

        return MapEntry(key, value.toString());
      },
    );
  }
}
